﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace DSASearch {
    class CustomHashTable {
        readonly Int32[] _array;
        readonly List<Int32> _probeArray;
        readonly Int32 _bucketSize;

        public CustomHashTable(Int32 size, Double loadfactor) {
            // ensure if minimal array size is not less than 3.
            size = Math.Max(3, size);
            // calculate internal array extra size. Worst value is 0, best value is 0.3
            // values greater than 0.3 do not improve performance, but waste the memory
            // at least one additional cell is added for efficient search purposes
            Int32 extraMemory = (Int32) Math.Max(1, size * loadfactor);
            _bucketSize = getNextPrime(size + extraMemory);
            _array = new Int32[_bucketSize];
            // initialize internal array that will be used for statistics purposes and will
            // hold number of probes performed during value insertion in each addition operation
            // value of 1 means "no collisions", any higher value indicates the number of
            // collisions detected
            _probeArray = new List<Int32>();
            // set public property to indicate the accessible size of the hashtable.
            Capacity = size;
        }

        /// <summary>
        /// Gets the public capacity of the hashtable. It doesn't reflect internal array size
        /// if the load factor is non-zero.
        /// </summary>
        public Int32 Capacity { get; }
        /// <summary>
        /// Gets the number of elements stored in the hashtable. If the value is equals to
        /// <see cref="Capacity"/>, the hashtable is full and no new elements could be added.
        /// </summary>
        public Int32 Length { get; private set; }

        /// <summary>
        /// Adds new element to hashtable.
        /// </summary>
        /// <param name="value">A value to add</param>
        /// <returns>A total number of comparisons were made during empty slot search.</returns>
        public Int32 Add(Int32 value) {
            if (Capacity == Length) { throw new OverflowException("HashTable is full"); }
            Int32 insertprobes;
            Int32 slot = getInsertSlotIndex(value, out insertprobes);
            _array[slot] = value;
            Length++;
            return insertprobes * 2;
        }
        /// <summary>
        /// Finds the key in the hashtable.
        /// </summary>
        /// <param name="key">Key to find</param>
        /// <returns>A total number of comparisons were made during key search.</returns>
        public Int32 FindKey(Int32 key) {
            Int32 findprobes;
            var index = getFindSlotIndex(key, out findprobes);
            return findprobes * 2 + 1;
        }
	    public Int32 DeleteKey(Int32 key) {
			Int32 findprobes;
			var index = getFindSlotIndex(key, out findprobes);
			_array[index] = -1;
			return findprobes * 2 + 1;
		}
        public void Print() {
            for (Int32 i = 0; i < _array.Length - 1; i++) {
                Console.Write(_array[i] + " ");
            }
        }
        /// <summary>
        /// Returns the statistics information. It is an array of probes made during each insertion
        /// operation.
        /// </summary>
        /// <returns>Statistics information</returns>
        public List<Int32> GetStatistics() {
            return _probeArray.ToList();
        }
        // gets the slot index where new element will be stored.
        Int32 getInsertSlotIndex(Int32 value, out Int32 insertprobes) {
            Int32 index;
            insertprobes = 0;
            do {
                insertprobes++;
                index = getHash(value, insertprobes);
            } while (_array[index] > 0 && _array[index] != value);
            if (_array[index] == value) { throw new ArgumentException("The key already exists."); }
            _probeArray.Add(insertprobes);
            return index;
        }
        // gets the slot index where the requested key is stored. If key is not stored in the hashtable,
        // the method returns -1.
        Int32 getFindSlotIndex(Int32 key, out Int32 findprobes) {
            Int32 index;
            findprobes = 0;
            do {
                findprobes++;
                index = getHash(key, findprobes);
            } while (_array[index] != key && _array[index] != 0);
            return _array[index] == key ? index : -1;
        }
        // calculates the hash for a given value and multiplicative factor if rehashing is required.
        Int32 getHash(Int32 value, Int32 factor) {
            return Math.Abs((value + factor * (1 + ((value >> 5) + 1) % (_bucketSize - 1))) % _bucketSize);
        }
        // gets nearest prime number which is larger than the given value. Due to specific
        // implementation, internal array size is tied (aligned) to primes.
        static Int32 getNextPrime(Int32 baseValue) {
            while (!isPrime(baseValue)) {
                baseValue++;
            }
            return baseValue;
        }
        // improved static method to determine if a given number is prime or not.
		static Boolean isPrime(Int32 number) {
			Int32 boundary = (Int32)Math.Floor(Math.Sqrt(number));
			if (number == 1) { return false; }
			if (number == 2) { return true; }
			for (Int32 i = 3; i <= boundary; i += 2) {
				if (number % i == 0) { return false; }
			}
			return true;
		}
    }
}
